<?php

namespace EasyContent\WordPress;

use EasyContent\WordPress\Api\Models\Article;
use EasyContent\WordPress\Db\Models\Article as ArticleDbModel;
use EasyContent\WordPress\Db\Models\ArticlePost;
use EasyContent\WordPress\Db\Models\WorkflowStage;
use EasyContent\WordPress\Exceptions\EasyContentResponseException;
use EasyContent\WordPress\Helpers\ArticleWorkflowStageUpdater;
use EasyContent\WordPress\Helpers\CategoriesMapper;
use EasyContent\WordPress\Helpers\PostContentProcessor;
use EasyContent\WordPress\Helpers\TimeHelper;
use GuzzleHttp\Exception\BadResponseException;

final class Post
{

    /** @var null|\WP_Post $_wpPostObject Original WordPress post object */
    protected $_wpPostObject;

    /** @var null|integer $_synchronizationTime */
    protected $_synchronizationTime = false;

    /** @var null|ArticleDbModel $_dbArticleObject */
    protected $_dbArticleObject = false;


    /**
     * @param int|\WP_Post $post
     * @throws \Exception
     */
    public function __construct($post)
    {
        $postObject = $post instanceof \WP_Post ? $post : null;

        if ( is_int( $post ) ) {
            if ( $post <= 0 ) {
                throw new \Exception( __( 'Invalid post ID', EASYCONTENT_TXTDOMAIN ), 1 );
            }

            $postObject = get_post( $post );
        }

        if ( null === $postObject ) {
            throw new \Exception( __( 'Error during post fetching', EASYCONTENT_TXTDOMAIN ), 2 );
        }

        $validTypes = ['post', 'page'];

        if ( ! in_array( $postObject->post_type, $validTypes ) ) {
            throw new \Exception( __( 'Invalid post type', EASYCONTENT_TXTDOMAIN ), 3 );
        }

        $this->_wpPostObject = $postObject;
    }


    public function getId()
    {
        return $this->_wpPostObject->ID;
    }


    /**
     * Gets article object based on cached DB data
     *
     * @return ArticleDbModel|null
     */
    public function getArticle()
    {
        if ( false === $this->_dbArticleObject ) {
            $this->loadArticle();
        }

        return $this->_dbArticleObject;
    }


    protected function loadArticle()
    {
        $this->_dbArticleObject = ArticleDbModel::getByPostId( $this->_wpPostObject->ID );
    }


    /**
     * @return int|null Returns UNIX timestamp or NULL if nothing set
     */
    public function getSynchronizationTime()
    {
        if ( false === $this->_synchronizationTime ) {
            $this->_synchronizationTime = $this->isSynced()
                ? ArticlePost::getSynchronizationTime( $this->getArticle()->id, $this->_wpPostObject->ID )
                : null;
        }

        return $this->_synchronizationTime;
    }


    public function getSynchronizationTimeString()
    {
        $time = $this->getSynchronizationTime();

        if ( null === $time ) {
            return ''; // &mdash;
        }

        return TimeHelper::getArticleListTableDateString( $time );
    }


    public function isSynced()
    {
        return (bool)$this->getArticle();
    }


    /**
     * Pulls latest article revision from EasyContent.io
     *
     * @throws \Exception
     */
    public function pull()
    {
        if ( ! $this->isSynced() ) {
            throw new \Exception( __( 'Post was not synchronized. Push first!', EASYCONTENT_TXTDOMAIN ) );
        }

        // Gets data from remote host (EasyContent.io)
        // This line will throw exception on bad request
        $article = Article::findOne( $this->getArticle()->id );

        if ( null === $article ) {
            // Article was not found on remote host
            // In this case we need to remove it from WP DB

            ArticlePost::detach( $this->getArticle()->id, $this->_wpPostObject->ID );
            ArticleDbModel::deleteByPrimaryKey( $this->getArticle()->id );

            $this->_dbArticleObject = null;

            throw new \Exception( __( 'Unable to find article on EasyContent.io', EASYCONTENT_TXTDOMAIN ), 10 );
        }

        $postData = array_merge( get_object_vars( $this->_wpPostObject ), [
            'post_title'   => $article->title,
            'post_excerpt' => $article->excerpt ?: '',
            'post_name'    => $article->custom_slug ?: ''
        ] );

        $postId = wp_update_post( wp_slash( $postData ) );

        if ( 0 === $postId ) {
            throw new \Exception( __( 'Error during saving remote post data', EASYCONTENT_TXTDOMAIN ) );
        }

        $this->_wpPostObject = get_post( $postId );

        $this->setContent( $article->content );
        $this->setFeaturedImageByUrl( $article->featured_image );

        wp_set_post_tags( $postId, $article->tags );

        if ( null !== $article->meta_title ) {
            $this->updateMeta( ['meta_title' => (string)$article->meta_title] );
        }

        if ( null !== $article->meta_description ) {
            $this->updateMeta( ['meta_description' => (string)$article->meta_description] );
        }

        $articleAttributes = $article->getAttributes();

        $updateResult = ArticleDbModel::update(
            $articleAttributes,
            ['id' => $this->getArticle()->id],
            ['%d', '%d', '%s', '%s', '%d'],
            ['%d']
        );

        if ( ! $updateResult ) {
            throw new \Exception( __( 'Error during cache update', EASYCONTENT_TXTDOMAIN ) );
        }

        $this->loadArticle();
        ArticlePost::setSynchronizationTime( $this->getArticle()->id, $this->_wpPostObject->ID );

        return $this->getFrontendData();
    }


    /**
     * Pushes article from WordPress site to EasyContent.io
     */
    public function push()
    {
        try {
            $parameters = [
                'title' => $this->_wpPostObject->post_title,
                'content' => $this->_wpPostObject->post_content,
                'meta_title' => $this->getMetaTitle(),
                'meta_description' => $this->getMetaDescription(),
                'featured_image' => $this->getFeaturedImageUrl(),
                'excerpt' => $this->_wpPostObject->post_excerpt,
                'tags' => $this->getPostTagsNames(),
                'custom_post_slug' => $this->getPostSlug()
            ];

            if ( ! $this->isSynced() ) {
                $parameters['categories'] = CategoriesMapper::getRelatedCategoriesList( $this->_wpPostObject );
            }

            // Creates API model to perform requests
            $article = ! $this->isSynced()
                ? Article::create( $parameters )
                : ( new Article( $this->getArticle()->id ) )->addNewRevision( $parameters );


            $articleAttributes = $article->getAttributes();

            if ( ! $this->isSynced() ) {
                // In this case we don't have data in DB so need to add one

                $newArticleDbModel = ArticleDbModel::insertRow(
                    $articleAttributes,
                    ['%d', '%d', '%s', '%s', '%d']
                );

                if ( ! $newArticleDbModel ) {
                    // Error during inserting article
                    // In this case data was pushed but DB cache was not updated

                    throw new \Exception( __( 'Error during cache update', EASYCONTENT_TXTDOMAIN ), 13 );
                }

                $this->_dbArticleObject = $newArticleDbModel;

                ArticlePost::attach( $this->getArticle()->id, $this->_wpPostObject->ID );
            } else {
                // We have article data in DB and we need to update it

                $updateResult = ArticleDbModel::update(
                    $articleAttributes,
                    ['id' => $this->getArticle()->id],
                    ['%d', '%d', '%s', '%s', '%d'],
                    ['%d']
                );

                if ( ! $updateResult ) {
                    throw new \Exception( __( 'Error during cache update', EASYCONTENT_TXTDOMAIN ), 14 );
                }

                $this->loadArticle();
                ArticlePost::setSynchronizationTime( $this->getArticle()->id, $this->_wpPostObject->ID );
            }

            // We are pushing WP post to EC so no need to update post_content property

            return $this->getFrontendData();
        } catch ( BadResponseException $e ) {
            $ecException = new EasyContentResponseException( $e );

            if ( $ecException->getCode() == EasyContentResponseException::ARTICLE_NOT_FOUND_CODE ) {
                // Post was synchronized but remote article was not found
                ArticleDbModel::deleteByPrimaryKey( $this->getArticle()->id );
                ArticlePost::detach( $this->getArticle()->id, $this->_wpPostObject->ID );

                $this->_dbArticleObject = null;
            }

            throw new \Exception( $ecException->getMessage(), 12 );
        }
    }


    public function getFrontendData()
    {
        if ( ! $this->isSynced() ) {
            return null;
        }

        $article = $this->getArticle();
        $currentWorkflowStage = $article->getCurrentWorkflowStage();

        return [
            'id'        => $article->id, // Remote article ID. Used for endpoints generation
            'isSynced'  => true,
            'title'     => $article->title,
            'topicName' => $article->topic_name,
            'url'       => $article->getEditRemoteArticleUrl(),
            'content'   => $this->_wpPostObject->post_content, // This probably doesn't work. TODO: Fix this and related quickfixes
            'stage'     => $currentWorkflowStage ? [
                'id'    => $currentWorkflowStage->id,
                'name'  => $currentWorkflowStage->name,
                'color' => $currentWorkflowStage->color
            ] : null,
            'lastUpdatedString' => $this->getSynchronizationTimeString()
        ];
    }


    public static function createFromArticleId($articleId)
    {

        $junction = ArticlePost::findByKey( $articleId, 'article_id' );

        return $junction ? new static( (int)$junction->post_id ) : null;
    }


    public function updateMeta(array $params)
    {
        $seoPlugin = Plugin::getInstance()->getSettings()->getSeoPlugin();

        if ( $seoPlugin && isset( $params['meta_title'] ) ) {
            $titleMetaKey = 'yoast-seo' === $seoPlugin ? '_yoast_wpseo_title' : '_aioseop_title';
            update_post_meta( $this->_wpPostObject->ID, $titleMetaKey, $params['meta_title'] );
        }

        if ( $seoPlugin && isset( $params['meta_description'] ) ) {
            $descriptionMetaKey = 'yoast-seo' === $seoPlugin ? '_yoast_wpseo_metadesc' : '_aioseop_description';
            update_post_meta( $this->_wpPostObject->ID, $descriptionMetaKey, $params['meta_description'] );
        }
    }


    public function getMetaTitle()
    {
        $seoPlugin = Plugin::getInstance()->getSettings()->getSeoPlugin();

        if ( ! $seoPlugin ) {
            return null;
        }

        $metaKey = 'yoast-seo' === $seoPlugin ? '_yoast_wpseo_title' : '_aioseop_title';

        return get_post_meta( $this->_wpPostObject->ID, $metaKey, true );
    }


    public function getMetaDescription()
    {
        $seoPlugin = Plugin::getInstance()->getSettings()->getSeoPlugin();

        if ( ! $seoPlugin ) {
            return null;
        }

        $metaKey = 'yoast-seo' === $seoPlugin ? '_yoast_wpseo_metadesc' : '_aioseop_description';

        return get_post_meta( $this->_wpPostObject->ID, $metaKey, true );
    }


    public function getFeaturedImageUrl()
    {
        return get_the_post_thumbnail_url( $this->_wpPostObject, 'full' ) ?: '';
    }


    public function getPostSlug()
    {
        return get_post_field( 'post_name', $this->_wpPostObject );
    }


    public function getPostTagsNames()
    {
        /** @var \WP_Term[]|false|\WP_Error */
        $tags = get_the_tags( $this->_wpPostObject->ID );

        if ( false === $tags || is_wp_error( $tags ) ) {
            return [];
        }

        $strings = [];

        /** @var \WP_Term $tag */
        foreach ( $tags as $tag ) {
            $strings[] = $tag->name;
        }

        return $strings;
    }


    public function setContent($content = '')
    {
        $postContent = $content
            ? ( new PostContentProcessor( $content, $this->_wpPostObject->ID ) )->run()
            : '';

        $result = wp_update_post( [
            'ID' => $this->_wpPostObject->ID,
            'post_content' => $postContent
        ], true );

        if ( ! is_wp_error( $result ) ) {
            $this->_wpPostObject->post_content = $postContent;
        }

        return ! is_wp_error( $result );
    }


    public function setFeaturedImageByUrl($url)
    {
        if ( ! is_string( $url ) ) {
            return;
        }

        $attachment = Attachment::createFromUrl( $url, $this->_wpPostObject->ID );

        if ( ! $attachment ) {
            return;
        }

        set_post_thumbnail( $this->_wpPostObject, $attachment->getPostObject()->ID );
    }


    public static function onDelete($postId)
    {
        try {
            $post = new Post( (int)$postId ); // Used to validate postId and other things only
            ArticlePost::deleteByPrimaryKey( $postId, 'post_id' );
        } catch ( \Exception $e ) {
        }
    }


    public static function onSave($postId, $post, $update)
    {
        try {
            $p = new self( $post );

            $newWorkflowStageId = isset( $_POST['ec_new_article_ws_id'] ) && isset( $_POST['ec_change_article_ws'] )
                ? absint( $_POST['ec_new_article_ws_id'] )
                : 0;

            if ( $newWorkflowStageId > 0
                && ( $article = $p->getArticle() )
                && ( $currentWorkflowStage = $article->getCurrentWorkflowStage() )
                && $currentWorkflowStage->id != $newWorkflowStageId
                && ( $workflowStage = WorkflowStage::findByKey( $newWorkflowStageId ) )
            ) {
                $result = ( new ArticleWorkflowStageUpdater() )->update( $p, $newWorkflowStageId );

                if ( is_wp_error( $result ) ) {
                    // TODO: Add error notice
                }
            }


            $pullPost = isset( $_POST['ec_pull_post'] );
            $pushPost = isset( $_POST['ec_push_post'] );

            // Unset variables to prevent multiple execution onPostPull
            unset( $_POST['ec_pull_post'], $_POST['ec_push_post'] );

            if ( ! $pullPost && ! $pushPost ) {
                return;
            }

            $pullPost ? $p->pull() : $p->push();
        } catch ( \Exception $e ) {
            // TODO: add notice
        }
    }
}
