<?php

namespace EasyContent\WordPress\Api\Models;

use EasyContent\WordPress\Db\Models\ArticlePost;
use EasyContent\WordPress\Db\Models\WorkflowStage;
use EasyContent\WordPress\Helpers\CategoriesMapper;
use EasyContent\WordPress\Plugin;
use EasyContent\WordPress\Post;
use EasyContent\WordPress\Db\Models\Article as ArticleDBModel;

/**
 * @property integer $id
 * @property integer $current_workflow_stage_id
 * @property int[] $categories
 * @property string $topic_name
 * @property null|string $title
 * @property null|integer $updated
 * @property string $content
 * @property null|string $meta_title
 * @property null|string $meta_description
 * @property null|string $featured_image
 * @property null|string $excerpt
 * @property array|string[] $tags
 * @property null|string $custom_slug
 */
final class Article extends BaseModel
{
    private $_articlePostDbRelation = false;

    public $id;
    public $current_workflow_stage_id;
    public $categories;
    public $topic_name;
    public $title;
    public $updated;
    public $content;
    public $meta_title;
    public $meta_description;
    public $featured_image;
    public $excerpt;
    public $tags;
    public $custom_slug;


    public function __construct($remoteArticleId = null)
    {
        $this->id = $remoteArticleId;
    }


    protected static function getDefaults()
    {
        return [
            'id'                        => 0,
            'categories'                => [],
            'current_workflow_stage_id' => 0,
            'title'                     => __( 'N/A', EASYCONTENT_TXTDOMAIN ),
            'topic_name'                => __( 'N/A', EASYCONTENT_TXTDOMAIN ),
            'updated'                   => 0,
            'content'                   => '',
            'meta_title'                => '',
            'meta_description'          => '',
            'featured_image'            => '',
            'excerpt'                   => '',
            'tags'                      => [],
            'custom_slug'               => ''
        ];
    }


    protected static function getFields()
    {
        return ['id', 'current_workflow_stage_id', 'categories',
            'topic_name', 'title', 'updated', 'content', 'meta_title', 'meta_description',
            'featured_image', 'excerpt', 'tags', 'custom_slug'
        ];
    }


    public static function getEndpoint()
    {
        return 'articles';
    }


    public static function create(array $params)
    {
        $plugin = Plugin::getInstance();

        $plugin->log( 'Creating article on EasyContent.io with POST request', $params );

        return static::makeRequest( 'POST', static::getEndpoint(), ['form_params' => $params] );
    }


    public function addNewRevision(array $params)
    {
        $plugin = Plugin::getInstance();

        $plugin->log( 'Creating new article revision with POST request', [
            'articleAttributes' => $this->getAttributes(),
            'params' => $params
        ] );

        $endpoint = static::getEndpoint();
        $endpoint = "{$endpoint}/{$this->id}/revisions";

        return static::makeRequest( 'POST', $endpoint, ['form_params' => $params] );
    }


    public function updateWorkflowStage($workflowStageId)
    {
        $endpoint = static::getEndpoint();
        $endpoint = "{$endpoint}/{$this->id}";

        return static::makeRequest( 'PATCH', $endpoint, [
            'form_params' => [
                'workflow_stage_id' => $workflowStageId
            ]
        ] );
    }


    /**
     * @param null|WorkflowStage $newStage
     * @return array|\WP_Error|Post
     */
    public function import(WorkflowStage $newStage = null)
    {
        $parameters = $this->getInsertingPostParameters();

        $postId = wp_insert_post( $parameters, true );

        if ( is_wp_error( $postId ) || 0 === $postId ) {
            // Error occurred. Post was not created/updated
            return new \WP_Error( $postId->get_error_code(), $postId->get_error_message() );
        }

        $post = new Post( (int)$postId );

        $post->setContent( $this->content );
        $post->setFeaturedImageByUrl( $this->featured_image );
        $post->updateMeta( [
            'meta_title' => $this->meta_title,
            'meta_description' => $this->meta_description
        ] );

        wp_set_post_tags( $postId, $this->tags );

        $this->updateCacheAfterImport( $postId, $newStage );

        return $post;
    }


    /**
     * @return ArticlePost|null
     */
    private function getArticlePostDbRelation()
    {
        if ( false === $this->_articlePostDbRelation ) {
            try {
                $this->_articlePostDbRelation = ArticlePost::findByKey( $this->id, 'article_id' );

                if ( null !== $this->_articlePostDbRelation ) {
                    new Post( (int)$this->_articlePostDbRelation->post_id );
                }
            } catch ( \Exception $e ) {
                $this->_articlePostDbRelation = null;
            }
        }

        return $this->_articlePostDbRelation;
    }


    private function getInsertingPostParameters()
    {
        $plugin = Plugin::getInstance();

        $parameters = [
            'post_author'    => $plugin->getSettings()->getPostAuthorId(),
            'post_status'    => $plugin->getSettings()->postStatus,
            'post_type'      => $plugin->getSettings()->postType,
            'comment_status' => $plugin->getSettings()->commentsStatus,
            'post_title'     => $this->title,
            'post_excerpt'   => $this->excerpt ?: '',
            'post_name'      => $this->custom_slug ?: '',
            'post_content'   => ''
        ];

        // Sets related post ID if article and post were connected
        $articlePostRelation = $this->getArticlePostDbRelation();

        if ( $articlePostRelation ) {
            $parameters['ID'] = $articlePostRelation->post_id;
        }

        // Sets post's category according to mapper rules
        if ( 'post' === $parameters['post_type'] ) {
            $parameters['post_category'] = CategoriesMapper::getWpCategoriesIds( $this->categories );
        }

        return $parameters;
    }


    /**
     * @param int $postId
     * @param null|WorkflowStage $stage
     */
    private function updateCacheAfterImport($postId, WorkflowStage $stage = null)
    {
        $newStageId = $this->current_workflow_stage_id;

        if ( $stage && $stage->id != $this->current_workflow_stage_id ) {
            try {
                $this->updateWorkflowStage( $stage->id );
                $newStageId = $stage->id;
            } catch ( \Exception $e ) { }
        }

        $fields = [
            'current_workflow_stage_id' => $newStageId,
            'topic_name'  => $this->topic_name,
            'title'       => $this->title,
            'updated'     => time()
        ];

        $cachedArticle = null;

        try {
            $cachedArticle = ArticleDBModel::findByKey( $this->id );
        } catch ( \Exception $e ) { }

        if ( $cachedArticle ) {
            ArticleDBModel::update(
                $fields,
                ['id' => $cachedArticle->id],
                ['%d', '%s', '%s', '%d'],
                ['%d']
            );
        } else {
            $fields['id'] = $this->id;
            ArticleDBModel::insertRow( $fields, ['%d', '%s', '%s', '%d', '%d'] );
        }


        if ( ! $this->getArticlePostDbRelation() ) {
            ArticlePost::insertRow( [
                'article_id' => $this->id,
                'post_id' => $postId,
                'sync_time' => time()
            ], ['%d', '%d', '%d'] );
        } else {
            ArticlePost::update(
                ['sync_time' => time()],
                [
                    'article_id' => $this->id,
                    'post_id' => $postId
                ],
                ['%d'],
                ['%d', '%d']
            );
        }
    }


    public function getAttributes(array $attributeNames = [])
    {
        $attributes = parent::getAttributes( $attributeNames );

        unset( $attributes['categories'] );

        return $attributes;
    }
}
