<?php

namespace EasyContent\WordPress\Api\Models;

use EasyContent\WordPress\Api\Client;

abstract class BaseModel
{
    abstract protected static function getFields();


    protected static function getDefaults()
    {
        return [];
    }


    abstract public static function getEndpoint();


    public static function findAll($throwException = false)
    {
        try {
            return static::makeRequest( 'GET', static::getEndpoint() );
        } catch (\Exception $e) {
            if ( $throwException ) {
                throw $e;
            }

            return [];
        }
    }


    /**
     * @param int $id ID of model to retrieve
     *
     * @throws \Exception
     *
     * @return null|static NULL if no model was found or model instance otherwise
     */
    public static function findOne($id)
    {
        $id = (int)$id;

        if ( $id <= 0 ) {
            return null;
        }

        $endpoint = static::getEndpoint() . "/{$id}";

        return static::makeRequest( 'GET', $endpoint );
    }


    protected static function makeRequest($method, $endpoint, $options = [])
    {
        $client = Client::create();
        $response = $client->request( $method, $endpoint, $options );

        if ( ! static::validateResponseCode( $method, $response->getStatusCode() ) ) {
            throw new \Exception( $response->getReasonPhrase(), 7 );
        }

        $data = json_decode( $response->getBody()->getContents() );

        if ( is_object( $data ) ) {
            return static::createModel( (array)$data );
        } elseif ( is_array( $data ) ) {
            return static::createList( $data );
        }

        throw new \Exception( __( 'Data decoding error', EASYCONTENT_TXTDOMAIN ), 8 );
    }


    protected static function validateResponseCode($method, $code)
    {
        $requestMethodValidStatus = [
            'GET'    => 200,
            'HEAD'   => 200,
            'POST'   => 201,
            'PUT'    => 200,
            'PATCH'  => 200,
            'DELETE' => 204
        ];

        return $requestMethodValidStatus[$method] == $code;
    }


    /**
     * @param array $data
     * @return static
     */
    private static function createModel(array $data)
    {
        $model = new static();
        $defaults = static::getDefaults();

        foreach ( static::getFields() as $mapping => $fieldName ) {
            $dataFieldName = is_int( $mapping ) ? $fieldName : $mapping;
            $defaultValue = isset( $defaults[$fieldName] ) ? $defaults[$fieldName] : null;

            $model->$fieldName = isset( $data[$dataFieldName] ) ? $data[$dataFieldName] : $defaultValue;
        }

        return $model;
    }


    /**
     * @param array $data Array of objects of data
     *
     * @return static[]
     */
    public static function createList(array $data)
    {
        $list = [];

        foreach ( $data as $row ) {
            $list[] = static::createModel( (array)$row );
        }

        return $list;
    }


    public function getAttributes(array $attributeNames = [])
    {
        $attributes = [];

        foreach ( static::getFields() as $mapping => $fieldName ) {
            if ( $attributeNames ) {
                if ( in_array( $fieldName, $attributeNames ) ) {
                    $attributes[$fieldName] = $this->getFieldValue( $fieldName );
                }
            } else {
                $attributes[$fieldName] = $this->getFieldValue( $fieldName );
            }
        }

        return $attributes;
    }


    public function getFieldValue($fieldName, $defaultValue = null)
    {
        $defaults = static::getDefaults();

        return property_exists( $this, $fieldName )
            ? $this->$fieldName
            : ( null !== $defaultValue
                ? $defaultValue
                : ( isset( $defaults[$fieldName] )
                    ? $defaults[$fieldName]
                    : null
                )
            );
    }
}
