<?php

namespace EasyContent\WordPress\Db;

use EasyContent\WordPress\Db\Models\Article;
use EasyContent\WordPress\Db\Models\ArticlePost;
use EasyContent\WordPress\Db\Models\Category;
use EasyContent\WordPress\Db\Models\WorkflowStage;
use EasyContent\WordPress\Db\Phinx\Application as PhinxApplication;
use EasyContent\WordPress\Db\Phinx\Wrapper\TextWrapper;
use EasyContent\WordPress\Plugin;

class DbController
{
    const DB_VERSION_OPTION_NAME = 'ec-current-db-version';
    const DEFAULT_DB_VERSION = '0.0.0';

    /** @var Plugin $plugin */
    private $plugin;


    public function __construct()
    {
        $this->plugin = Plugin::getInstance();

        $this->plugin->log( 'Creating database controller...' );
    }


    public function checkDbVersion()
    {
        try {
            if ( ! $this->needMigration() ) {
                return;
            }
        } catch ( \Exception $e ) {
            return;
        }

//        $message = ( new View() )->render( 'admin/notices/db-is-not-up-to-date', [
//            'currentVersion' => $this->getCurrentDatabaseVersion(),
//            'newVersion' => EASYCONTENT_DB_VERSION
//        ] );
//
//        $this->plugin->noticesManager->addNotice( $message, 'error' );
    }


    public function migrate()
    {
        $this->plugin->log( 'Starting migration...' );

        try {
            if ( ! $this->needMigration() ) {
                $this->plugin->log( 'DB is up to date. No migration will be performed' );

                return true;
            }
        } catch ( \Exception $e ) {
            $this->plugin->log( 'Migration error: ' . $e->getMessage() );

            return false;
        }

        $phinxApp = new PhinxApplication();
        $wrapper = new TextWrapper( $phinxApp );

        $this->plugin->log( 'Executing migration command...' );

        $result = $wrapper->getMigrate( 'production' ); // TODO: Add version

        $this->plugin->log( 'Output: ' . $result );

        if ( 0 !== $wrapper->getExitCode() ) {
            $this->plugin->log( 'Exit code is not zero. Error occurred...' );

            return false;
        }

        try {
            $this->plugin->log( 'Updating DB version WP option' );

            $this->setDbVersionToCurrent();
        } catch ( \Exception $e ) {
            $this->plugin->log( $e->getMessage() );
        }

        return true;
    }


    protected function needMigration()
    {
        $currentDbVersion = $this->getCurrentDatabaseVersion();

        $comparisonResult = version_compare( EASYCONTENT_DB_VERSION, $currentDbVersion );

        $this->plugin->log( 'Checking if migration is needed...', [
            'currentDbVersion' => $currentDbVersion,
            'pluginDbVersion'  => EASYCONTENT_DB_VERSION
        ] );

        // If versions are equal, no need to perform migration
        // Perform migrations in other cases only (when result is 1 or -1)

        return 0 !== $comparisonResult;
    }


    protected function getCurrentDatabaseVersion()
    {
        $version = (string)get_option( self::DB_VERSION_OPTION_NAME, self::DEFAULT_DB_VERSION );
        $pattern = '/^(([1-9]\d*)|0)\.(([1-9]\d*)|0)\.(([1-9]\d*)|0)$/m';

        if ( 1 !== preg_match( $pattern, $version ) ) {
            // Saved DB version value is invalid or missing

            $this->plugin->log( 'Saved DB version value is invalid or missing. Setting up default DB version' );

            // Setting invalid or missing option to default
            $version = self::DEFAULT_DB_VERSION;
            $updated = update_option( self::DB_VERSION_OPTION_NAME, $version, true );

            if ( ! $updated ) {
                throw new \Exception( __( 'Error occurred, please try again later', EASYCONTENT_TXTDOMAIN ) );
            }
        }

        return $version;
    }


    private function setDbVersionToCurrent()
    {
        $updated = update_option( self::DB_VERSION_OPTION_NAME, EASYCONTENT_DB_VERSION, true );

        if ( ! $updated ) {
            throw new \Exception( __( 'Error occurred, please try again later', EASYCONTENT_TXTDOMAIN ) );
        }

        return EASYCONTENT_DB_VERSION;
    }


    public function clearDatabase()
    {
        Article::drop();
        ArticlePost::drop();
        Category::drop();
        WorkflowStage::drop();

        global $wpdb;
        $tableName = $wpdb->prefix . BaseModel::EC_PLUGIN_DB_PREFIX . 'migration';
        $query = "DROP TABLE IF EXISTS `{$tableName}`";
        $wpdb->query( $query );

        delete_option( 'ec-current-db-version' );
        delete_option( 'ec-is-plugin-connected' );
        delete_option( 'ec-public-api-key' );
    }
}
